# -*- coding: utf-8 -*-
"""
Helper functions for gwsnr
"""

import os
import json

from importlib import resources
import pickle
import numpy as np
import bilby
from gwpy.timeseries import TimeSeries
# from scipy.interpolate import CubicSpline
from scipy.interpolate import interp1d


class NumpyEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for numpy data types. It converts numpy.ndarray objects (and any nested-list composition
    that includes ndarray objects) into regular lists for JSON serialization. This is particularly useful when
    serializing data structures that include numpy arrays.
    """

    def default(self, obj):
        # Check if the object is an instance of numpy.ndarray
        if isinstance(obj, np.ndarray):
            return obj.tolist()  # Convert ndarray to list
        # Fallback to default behavior for other types
        return super(NumpyEncoder, self).default(obj)

def save_json(file_name, param):
    """Save a dictionary as a json file.

    Parameters
    ----------
    file_name : `str`
        json file name for storing the parameters.
    param : `dict`
        dictionary to be saved as a json file.

    Example
    ----------
    >>> from ler.utils import save_json, load_json
    >>> param = dict_list = [{"name": "lens_mass", "type": "cubic_spline", "x": np.array([0, 1, 2, 3]), "y": [0, 1, 0, 1]},]
    >>> save_json("param.json", param)
    >>> param = load_json("param.json")
    """
    try:
        with open(file_name, "w", encoding="utf-8") as write_file:
            json.dump(param, write_file)
    except:
        with open(file_name, "w", encoding="utf-8") as write_file:
            json.dump(param, write_file, indent=4, cls=NumpyEncoder)

def load_json(file_name):
    """Load a json file.

    Parameters
    ----------
    file_name : `str`
        json file name for storing the parameters.

    Returns
    ----------
    param : `dict`
    """
    with open(file_name, "r", encoding="utf-8") as f:
        param = json.load(f)

    return param

def save_pickle(file_name, param):
    """Save a dictionary as a pickle file.

    Parameters
    ----------
    file_name : `str`
        pickle file name for storing the parameters.
    param : `dict`
        dictionary to be saved as a pickle file.
    """
    with open(file_name, "wb") as handle:
        pickle.dump(param, handle, protocol=pickle.HIGHEST_PROTOCOL)

def load_pickle(file_name):
    """Load a pickle file.

    Parameters
    ----------
    file_name : `str`
        pickle file name for storing the parameters.

    Returns
    ----------
    param : `dict`
    """
    with open(file_name, "rb") as f:
        param = pickle.load(f)

    return param

def load_ann_h5(filename):
    """
    Function to load a specific dataset from an .h5 file

    Parameters
    ----------
    filename : str
        name of the .h5 file
        
    Returns
    ----------
    model : `keras.models.Model`
        Keras model loaded from the .h5 file
    """
    # supress warning
    import absl.logging
    absl.logging.set_verbosity(absl.logging.ERROR)
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    from tensorflow.keras.models import load_model
    return load_model(filename)

def append_json(file_name, new_dictionary, old_dictionary=None, replace=False):
    """
    Append (values with corresponding keys) and update a json file with a dictionary. There are four options:

    1. If old_dictionary is provided, the values of the new dictionary will be appended to the old dictionary and save in the 'file_name' json file.
    2. If replace is True, replace the json file (with the 'file_name') content with the new_dictionary.
    3. If the file does not exist, create a new one with the new_dictionary.
    4. If none of the above, append the new dictionary to the content of the json file.

    Parameters
    ----------
    file_name : `str`
        json file name for storing the parameters. 
    new_dictionary : `dict`
        dictionary to be appended to the json file.
    old_dictionary : `dict`, optional
        If provided the values of the new dictionary will be appended to the old dictionary and save in the 'file_name' json file. 
        Default is None.
    replace : `bool`, optional
        If True, replace the json file with the dictionary. Default is False.

    """

    # check if the file exists
    # time
    # start = datetime.datetime.now()
    if old_dictionary:
        data = old_dictionary
    elif replace:
        data = new_dictionary
    elif not os.path.exists(file_name):
        #print(f" {file_name} file does not exist. Creating a new one...")
        replace = True
        data = new_dictionary
    else:
        #print("getting data from file")
        with open(file_name, "r", encoding="utf-8") as f:
            data = json.load(f)
    # end = datetime.datetime.now()
    # print(f"Time taken to load the json file: {end-start}")

    # start = datetime.datetime.now()
    if not replace:
        data = add_dictionaries_together(data, new_dictionary)
        # data_key = data.keys()
        # for key, value in new_dictionary.items():
        #     if key in data_key:
        #         data[key] = np.concatenate((data[key], value)).tolist()
    # end = datetime.datetime.now()
    # print(f"Time taken to append the dictionary: {end-start}")

    # save the dictionary
    # start = datetime.datetime.now()
    #print(data)
    with open(file_name, "w", encoding="utf-8") as write_file:
        json.dump(data, write_file, indent=4, cls=NumpyEncoder)
    # end = datetime.datetime.now()
    # print(f"Time taken to save the json file: {end-start}")

    return data

def add_dictionaries_together(dictionary1, dictionary2):
    """
    Adds two dictionaries with the same keys together.
    
    Parameters
    ----------
    dictionary1 : `dict`
        dictionary to be added.
    dictionary2 : `dict`
        dictionary to be added.

    Returns
    ----------
    dictionary : `dict`
        dictionary with added values.
    """
    dictionary = {}
    # Check if either dictionary empty, in which case only return the dictionary with values
    if len(dictionary1) == 0:
        return dictionary2
    elif len(dictionary2) == 0:
        return dictionary1
    # Check if the keys are the same
    if dictionary1.keys() != dictionary2.keys():
        raise ValueError("The dictionaries have different keys.")
    for key in dictionary1.keys():
        value1 = dictionary1[key]
        value2 = dictionary2[key]

        # check if the value is empty
        bool0 = len(value1) == 0 or len(value2) == 0
        # check if the value is an ndarray or a list
        bool1 = isinstance(value1, np.ndarray) and isinstance(value2, np.ndarray)
        bool2 = isinstance(value1, list) and isinstance(value2, list)
        bool3 = isinstance(value1, np.ndarray) and isinstance(value2, list)
        bool4 = isinstance(value1, list) and isinstance(value2, np.ndarray)
        bool4 = bool4 or bool3
        bool5 = isinstance(value1, dict) and isinstance(value2, dict)

        if bool0:
            if len(value1) == 0 and len(value2) == 0:
                dictionary[key] = np.array([])
            elif len(value1) != 0 and len(value2) == 0:
                dictionary[key] = np.array(value1)
            elif len(value1) == 0 and len(value2) != 0:
                dictionary[key] = np.array(value2)
        elif bool1:
            dictionary[key] = np.concatenate((value1, value2))
        elif bool2:
            dictionary[key] = value1 + value2
        elif bool4:
            dictionary[key] = np.concatenate((np.array(value1), np.array(value2)))
        elif bool5:
            dictionary[key] = add_dictionaries_together(
                dictionary1[key], dictionary2[key]
            )
        else:
            raise ValueError(
                "The dictionary contains an item which is neither an ndarray nor a dictionary."
            )
    return dictionary

def get_param_from_json(json_file):
    """
    Function to get the parameters from json file.

    Parameters
    ----------
    json_file : `str`
        json file name for storing the parameters.

    Returns
    ----------
    param : `dict`
    """
    with open(json_file, "r", encoding="utf-8") as f:
        param = json.load(f)

    for key, value in param.items():
        param[key] = np.array(value)
    return param

def load_ann_h5_from_module(package, directory, filename):
    """
    Function to load a specific dataset from an .h5 file within the package

    Parameters
    ----------
    package : str
        name of the package
    directory : str
        name of the directory within the package
    filename : str
        name of the .h5 file

    Returns
    ----------
    model : `keras.models.Model`
        Keras model loaded from the .h5 file
    """
    # supress warning
    import absl.logging
    absl.logging.set_verbosity(absl.logging.ERROR)
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    from tensorflow.keras.models import load_model

    with resources.path(package + '.' + directory, filename) as h5_path:
        return load_model(h5_path)

def load_json_from_module(package, directory, filename):
    """
    Function to load a specific dataset from a .json file within the package

    Parameters
    ----------
    package : str
        name of the package
    directory : str
        name of the directory within the package
    filename : str
        name of the .json file

    Returns
    ----------
    data : `dict`
        Dictionary loaded from the .json file
    """

    with resources.path(package + '.' + directory, filename) as json_path:
        with open(json_path, "r", encoding="utf-8") as f:
            return json.load(f)
    
def load_pickle_from_module(package, directory, filename):
    """
    Function to load a specific dataset from a .pkl file within the package

    Parameters
    ----------
    package : str
        name of the package
    directory : str
        name of the directory within the package
    filename : str
        name of the .pkl file

    Returns
    ----------
    data : `dict`
        Dictionary loaded from the .pkl file
    """
    
    with resources.path(package + '.' + directory, filename) as pkl_path:
        return pickle.load(open(pkl_path, "rb"))

def dealing_with_psds(psds=None, ifos=None, f_min=20.0, sampling_frequency=2048.0):
    """
    Function to deal with psds inputs and for creating bilby.gw.detector.PowerSpectralDensity objects.

    Parameters
    ----------
    psds : dict
        dictionary of psds. psds.keys()=detector names, psds.values()=psds file names or pycbc psd names
    ifos : `list` or `None`
        List of interferometer objects or interferometer name list. Default is None. If None, bilby's default interferometer objects will be used.
    f_min : `float`
        Minimum frequency of the psds. Default is 20.
    sampling_frequency : `float`
        Sampling frequency of the psds. Default is 2048.

    Returns
    ----------
    psds_list : `list`
        list of bilby.gw.detector.PowerSpectralDensity objects
    detector_tensor_list : `list`
        list of detector tensors
    detector_list : `list`
        list of detector names
    """

    if not psds and not ifos:
        # if psds is not given, choose bilby's default psds
        print("psds not given. Choosing bilby's default psds")
        psds = dict()
        psds["L1"] = "aLIGO_O4_high_asd.txt"
        psds["H1"] = "aLIGO_O4_high_asd.txt"
        psds["V1"] = "AdV_asd.txt"
        detector_list = list(psds.keys())
        # for Fp, Fc calculation
        ifos = bilby.gw.detector.InterferometerList(detector_list)

    elif ifos and not psds:
        ifos_ = []
        detector_list = []
        psds = dict()
        for ifo in ifos:
            if isinstance(ifo, str):
                if ifo == "ET":
                    ifos_ += bilby.gw.detector.InterferometerList([ifo])
                    detector_list.append("ET1")
                    detector_list.append("ET2")
                    detector_list.append("ET3")
                    psds["ET1"] = ifos_[-3].power_spectral_density.psd_file
                    psds["ET2"] = ifos_[-2].power_spectral_density.psd_file
                    psds["ET3"] = ifos_[-1].power_spectral_density.psd_file
                else:
                    ifos_.append(bilby.gw.detector.InterferometerList([ifo])[0])
                    detector_list.append(ifo)
                    psds[ifo] = ifos_[-1].power_spectral_density.psd_file
                    if not psds[ifo]:
                        psds[ifo] = ifos_[-1].power_spectral_density.asd_file

            else:
                ifos_.append(ifo)
                detector_list.append(ifo.name)
                psds[ifo.name] = ifo.power_spectral_density.psd_file
                if not psds[ifo.name]:
                    psds[ifo.name] = ifo.power_spectral_density.asd_file
        ifos = ifos_

    elif psds and not ifos:
        detector_list = list(psds.keys())
        ifos = bilby.gw.detector.InterferometerList(detector_list)

        if "ET" in detector_list:
            # insert ET1, ET2, ET3 inplace of ET in the detector_list
            idx = detector_list.index("ET")
            detector_list.pop(idx)
            detector_list.insert(idx, "ET1")
            detector_list.insert(idx + 1, "ET2")
            detector_list.insert(idx + 2, "ET3")

        # for i, ifo in enumerate(ifos):
        #     psds[ifo.name] = ifo.power_spectral_density.psd_file

    elif psds and ifos:
        detector_list = []
        ifos_ = []
        psds_ = dict()
        for ifo in ifos:
            if isinstance(ifo, str):
                if ifo == "ET":
                    ifos_ += bilby.gw.detector.InterferometerList([ifo])
                    detector_list.append("ET1")
                    detector_list.append("ET2")
                    detector_list.append("ET3")
                    psds_["ET1"] = psds["ET"]
                    psds_["ET2"] = psds["ET"]
                    psds_["ET3"] = psds["ET"]
                else:
                    ifos_.append(bilby.gw.detector.InterferometerList([ifo])[0])
                    detector_list.append(ifo)
                    psds_[ifo] = psds[ifo]

            else:
                ifos_.append(ifo)
                detector_list.append(ifo.name)
                psds_[ifo.name] = psds[ifo.name]

        ifos = ifos_
        psds = psds_
    else:
        raise ValueError("psds and ifos are not in the correct format")

    # generate bilby's psd objects
    psds_list = []
    detector_tensor_list = []
    error_msg = "the psds format is not recognised. The parameter psds dict should contain chosen detector names as keys and corresponding psds txt file name (or name from pycbc psd) as their values"
    # print(psds)
    # print(detector_list)
    for i, det in enumerate(detector_list):
        # either provided psd or what's available in bilby

        if type(psds[det]) == str and psds[det][-3:] == "txt":

            psds_list.append(
                # bilby.gw.detector.PowerSpectralDensity(asd_file=psds[det])
                power_spectral_density_bilby(psds[det])
            )

        elif isinstance(psds[det], float):
            # get the psd from the open data
            duration = 16
            sample_rate = 4096.0  # 2048Hz throws error
            psd_duration = duration * 32
            analysis_start = psds[det]
            psd_start_time = analysis_start - psd_duration

            # check directory
            if not os.path.exists("./psd_data"):
                os.makedirs("./psd_data")

            # check if the txt file exists in psd directory
            path_ = f'./psd_data/{det}_{int(analysis_start+1)}_psd.txt'
            it_exist = os.path.exists(path_)

            if not it_exist:
                # set up empty interferometer
                X1 = bilby.gw.detector.get_empty_interferometer(det)

                # download the data
                print(f"Downloading data for {det} detector")
                X1_psd_data = TimeSeries.fetch_open_data(
                    det, psd_start_time, psd_start_time + psd_duration, sample_rate=sample_rate, cache=True)

                # calculate the psd
                psd_alpha = 2 * X1.strain_data.roll_off / duration
                X1_psd = X1_psd_data.psd(fftlength=duration, overlap=0, window=("tukey", psd_alpha), method="median")

                # save the psd
                print(f"Saving psd data for {det} detector at {path_}")
                X1_psd.write(path_)

                # initialize the psd object
                psd_obj = bilby.gw.detector.PowerSpectralDensity(
                        frequency_array=X1_psd.frequencies.value, psd_array=X1_psd.value
                    )
            else:
                print(f"Loading psd data for {det} detector from {path_}")
                psd_obj = bilby.gw.detector.PowerSpectralDensity(psd_file=path_)

            psds_list.append(psd_obj)

        elif isinstance(psds[det], str):
            print("Trying to get the psd from pycbc: ", psds[det])
            psds_list.append(
                power_spectral_density_pycbc(psds[det], f_min, sampling_frequency)
            )
        else:
            raise ValueError(error_msg)
        
        detector_tensor_list.append(ifos[i].detector_tensor)

    return psds_list, detector_tensor_list, detector_list, ifos

def power_spectral_density_bilby(psd_txt):
    """
    psd array finder from bilby

    Parameters
    ----------
    psd_txt : str
        name of the psd
        e.g. 'aLIGOaLIGODesignSensitivityT1800044'

    Returns
    -------
    psd_array : bilby.gw.detector.psd.PowerSpectralDensity object
    """
    if psd_txt[-7:] == "asd.txt":
        psd_object = bilby.gw.detector.PowerSpectralDensity(asd_file=psd_txt)
    elif psd_txt[-7:] == "psd.txt":
        psd_object = bilby.gw.detector.PowerSpectralDensity(psd_file=psd_txt)
    else:
        raise ValueError(
            "psd file name should end with either 'psd.txt' or 'asd.txt'"
        )

    frequency_array = psd_object.frequency_array
    psd_array = psd_object.psd_array
    # spline_coeff = CubicSpline(frequency_array, psd_array).c
    spline_coeff = interp1d(frequency_array, psd_array, bounds_error=False, fill_value=np.inf)

    return [frequency_array, psd_array, spline_coeff]

def power_spectral_density_pycbc(psd, f_min=20.0, sampling_frequency=2048.0):
    """
    psd array finder from pycbc

    Parameters
    ----------
    psd : str
        name of the psd
        e.g. 'aLIGOaLIGODesignSensitivityT1800044'
    f_min : float
        minimum frequency of the psd
        default: 20.
    sampling_frequency : float
        sampling frequency of the psd
        default: 2048.

    Returns
    -------
    psd_array : bilby.gw.detector.psd.PowerSpectralDensity object
    """
    import pycbc
    import pycbc.psd

    delta_f = 1.0 / 16.0
    flen = int(sampling_frequency / delta_f)
    low_frequency_cutoff = f_min
    psd_ = pycbc.psd.from_string(psd, flen, delta_f, low_frequency_cutoff)
    frequency_array = psd_.get_sample_frequencies().data
    psd_array = psd_.to_frequencyseries().data
    # spline_coeff = CubicSpline(frequency_array, psd_array).c
    spline_coeff = interp1d(frequency_array, psd_array, bounds_error=False, fill_value=np.inf)

    return [frequency_array, psd_array, spline_coeff]


# interpolator check and generation
def interpolator_check(
    param_dict_given,
    interpolator_dir,
    create_new,
):
    """
    Function for interpolator (snr_partialsacaled) check and generation if not exists.

    Parameters
    ----------
    param_dict_given : dict
        dictionary of parameters based on which the existence of interpolator will be checked
    interpolator_dir : str
        path to the interpolator pickle file
    create_new : bool
        if True, new interpolator will be generated even if the interpolator exists
        if False, existing interpolator will be used if exists, otherwise new interpolator will be generated

    Returns
    ----------
    psds_list_ : list
        list of psd objects
    detector_tensor_list_ : list
        list of detector tensors
    detector_list_ : list
        list of detector names
    """

    detector_list = param_dict_given["detector"]
    psds_list = param_dict_given["psds"]
    detector_tensor_list = param_dict_given["detector_tensor"]
    detector_list_ = []
    detector_tensor_list_ = []
    psds_list_ = []
    path_interpolator_ = []
    path_interpolator_all = []

    # getting interpolator if exists
    # for each detector, one by one
    k = 0
    for det in detector_list:
        param_dict_given["detector"] = det
        param_dict_given["psds"] = psds_list[k]
        param_dict_given["detector_tensor"] = detector_tensor_list[k]
        # checking 
        path_interpolator, it_exist = interpolator_pickle_path(
            param_dict_given, interpolator_dir
        )
        if create_new:
            it_exist = False
        if it_exist:
            print(
                f"Interpolator will be loaded for {det} detector from {path_interpolator}"
            )
        else:
            print(
                f"Interpolator will be generated for {det} detector at {path_interpolator}"
            )

            detector_list_.append(det)
            psds_list_.append(psds_list[k])
            detector_tensor_list_.append(detector_tensor_list[k])
            path_interpolator_.append(path_interpolator)

        path_interpolator_all.append(path_interpolator)
        k += 1

    return (
        psds_list_,
        detector_tensor_list_,
        detector_list_,
        path_interpolator_,
        path_interpolator_all,
    )


def interpolator_pickle_path(param_dict_given, path="./interpolator_pickle"):
    """
    Function for storing or getting interpolator (snr_partialsacaled) pickle path

    Parameters
    ----------
    param_dict_given : dict
        dictionary of parameters based on which the existence of interpolator will be checked
    path : str
        path to the directory where the interpolator pickle file will be stored

    Returns
    -------
    path_interpolator : str
        path to the interpolator pickle file
        e.g. './interpolator_pickle/L1/partialSNR_dict_0.pickle'
    it_exist: bool
        True if the interpolator exists
        False if the interpolator does not exists
    """

    detector = param_dict_given["detector"]
    # arg no. for the detector from the detector list
    det_path = path + "/" + detector

    # check if path exists
    if not os.path.exists(path):
        os.makedirs(path)

    # check if detector path exists
    if not os.path.exists(det_path):
        os.makedirs(det_path)

    # check if param_dict_list.pickle exists
    if not os.path.exists(det_path + "/param_dict_list.pickle"):
        dict_list = []
        save_pickle(det_path + "/param_dict_list.pickle", dict_list)

    # for checking existing interpolator pickle in the det_path/param_dict_list.pickle file
    param_dict_stored = load_pickle(det_path + "/param_dict_list.pickle")

    len_ = len(param_dict_stored)
    # del param_dict_given["psds"]
    param_dict_given["psds"] = str(param_dict_given["psds"][0:2])
    param_dict_given["detector_tensor"] = str(param_dict_given["detector_tensor"])

    # # expilcitly check if the param_dict_given is not available in the param_dict_stored
    # for param_dict in param_dict_stored:
    #     for key in param_dict_given.keys():
    #         if param_dict_given[key] != param_dict[key]:
    #             print(f"param_dict_given {key} = {param_dict_given[key]} is not equal to param_dict_stored {key} = {param_dict[key]}")
    #         else:
    #             print(f"param_dict_given {key} = {param_dict_given[key]} is equal to param_dict_stored {key} = {param_dict[key]}")


    # print("\n\n", param_dict_given)
    # print("\n\n",param_dict_stored)
    if param_dict_given in param_dict_stored:
        # try and except is added so that user can regenerate a new interpolator pickle file just by
        # deleting the right file and reruing gwsnr with that params again
        # also, if the user delete the file by mistake, it will generate in the next run
        idx = param_dict_stored.index(param_dict_given)
        # check if interpolator pickle exists
        # get partialSNR interpolator if exists
        path_interpolator = det_path + "/partialSNR_dict_" + str(idx) + ".pickle"
        # there will be exception if the file is deleted by mistake
        if os.path.exists(path_interpolator):
            it_exist = True
        else:
            it_exist = False

    # if related dict not found in the param_dict_list.pickle
    else:
        it_exist = False
        path_interpolator = det_path + "/partialSNR_dict_" + str(len_) + ".pickle"
        # print("related dict not found in the param_dict_list.pickle, new interpolator will be generated")

        # store the pickle dict
        param_dict_stored.append(param_dict_given)
        save_pickle(det_path + "/param_dict_list.pickle", param_dict_stored)

    # print(f"In case if you need regeneration of interpolator of the given gwsnr param, please delete this file, {path_interpolator} \n")

    return (path_interpolator, it_exist)


def get_gw_parameters(gw_param_dict):

    if isinstance(gw_param_dict, dict):
        pass
    elif isinstance(gw_param_dict, str):
        try:
            gw_param_dict = get_param_from_json(gw_param_dict)
        except:
            raise ValueError("gw_param_dict should be either a dictionary or a json string")
    else:
        raise ValueError("gw_param_dict should be either a dictionary or a json string")
    
    # add error handling for all parameters
    for key, value in gw_param_dict.items():
        if isinstance(value, (list, np.ndarray)):
            if len(value) == 0 or np.any(np.isnan(value)) or np.any(np.isinf(value)):
                raise ValueError(f"Parameter '{key}' contains invalid values: {value}")
        elif value is None or np.isnan(value) or np.isinf(value):
            raise ValueError(f"Parameter '{key}' contains invalid value: {value}")

        if key in ["mass_1", "mass_2", "luminosity_distance", "geocent_time"]:
            if np.any(np.array(value) < 0):
                raise ValueError(f"Parameter '{key}' contains negative values: {value}")
            
        if key in ["a_1", "a_2", "eccentricity"]:
            if np.any(np.array(value) > 1):
                raise ValueError(f"Parameter '{key}' contains invalid values: {value}")

    mass_1 = gw_param_dict.get("mass_1")
    mass_2 = gw_param_dict.get("mass_2")
    mass_1, mass_2 = np.array([mass_1]).reshape(-1), np.array([mass_2]).reshape(-1)
    luminosity_distance = gw_param_dict.get("luminosity_distance", np.array([100.0]))
    theta_jn = gw_param_dict.get("theta_jn", np.array([0.0]))
    psi = gw_param_dict.get("psi", np.array([0.0]))
    phase = gw_param_dict.get("phase", np.array([0.0]))
    geocent_time = gw_param_dict.get("geocent_time", np.array([1246527224.169434]))
    ra = gw_param_dict.get("ra", np.array([0.0]))
    dec = gw_param_dict.get("dec", np.array([0.0]))
    size = len(mass_1)

    # Extract spin parameters or initialize to zeros
    a_1 = gw_param_dict.get("a_1", np.zeros(size))
    a_2 = gw_param_dict.get("a_2", np.zeros(size))

    # Extract precessing waveform parameters or initialize to zeros
    tilt_1 = gw_param_dict.get("tilt_1", np.zeros(size))
    tilt_2 = gw_param_dict.get("tilt_2", np.zeros(size))
    phi_12 = gw_param_dict.get("phi_12", np.zeros(size))
    phi_jl = gw_param_dict.get("phi_jl", np.zeros(size))

    # Extract tidal parameters or initialize to zeros
    lambda_1 = gw_param_dict.get("lambda_1", np.zeros(size))
    lambda_2 = gw_param_dict.get("lambda_2", np.zeros(size))
    
    # Extract eccentricity or initialize to zeros
    eccentricity = gw_param_dict.get("eccentricity", np.zeros(size))

    mass_1, mass_2 = np.array([mass_1]).reshape(-1), np.array([mass_2]).reshape(-1)
    (
        mass_1,
        mass_2,
        luminosity_distance,
        theta_jn,
        psi,
        phase,
        geocent_time,
        ra,
        dec,
        a_1,
        a_2,
        tilt_1,
        tilt_2,
        phi_12,
        phi_jl,
        lambda_1,
        lambda_2,
        eccentricity,
    ) = np.broadcast_arrays(
        mass_1,
        mass_2,
        luminosity_distance,
        theta_jn,
        psi,
        phase,
        geocent_time,
        ra,
        dec,
        a_1,
        a_2,
        tilt_1,
        tilt_2,
        phi_12,
        phi_jl,
        lambda_1,
        lambda_2,
        eccentricity,
    )

    return mass_1, mass_2, luminosity_distance, theta_jn, psi, phase, geocent_time, ra, dec, a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl, lambda_1, lambda_2, eccentricity


from .jaxjit_functions import (
    findchirp_chirptime_jax,
    antenna_response_plus,
    antenna_response_cross,
    antenna_response_array,
)

from .jaxjit_interpolators import (
    get_interpolated_snr_aligned_spins_jax,
    get_interpolated_snr_no_spins_jax,
)

# __all__ = [
#     # jaxjit_functions
#     'findchirp_chirptime_jax',
#     'antenna_response_plus',
#     'antenna_response_cross',
#     'antenna_response_array',
#     # jaxjit_interpolators
#     'get_interpolated_snr_aligned_spins_jax',
#     'get_interpolated_snr_no_spins_jax',
# ]





"""
GWSNR: Gravitational Wave Signal-to-Noise Ratio
"""
# In your package __init__.py

import os
import multiprocessing as mp
import warnings

def set_multiprocessing_start_method():
    """
    Set the multiprocessing start method based on OS and environment variables.
    By default, sets 'fork' on POSIX systems unless overridden by GWSNR_USE_SPAWN=True.
    """
    # Only set if not already set
    method = None

    # POSIX = Linux, Mac
    if os.name == 'posix':
        # User override: GWSNR_USE_SPAWN=True
        use_spawn = os.environ.get("GWSNR_USE_SPAWN", "False").lower() == "true"
        if use_spawn:
            method = "spawn"
            print("GWSNR: Setting multiprocessing start method to 'spawn' per environment variable.")
        else:
            method = "fork"
            # print(
            #     "GWSNR: Setting multiprocessing start method to 'fork'.\n"
            #     "If you need to use the 'spawn' method (in case error or warning due to other library dependencies),\n"
            #     "set the environment variable GWSNR_USE_SPAWN=True *before* running your script."
            #     "\n"
            #     "Command line (single line):\n"
            #     "    GWSNR_USE_SPAWN=True python yourscript.py\n"
            #     "In a Python script (before importing GWSNR):\n"
            #     "    import os\n"
            #     "    os.environ['GWSNR_USE_SPAWN'] = 'True'\n"
            #     "    import gwsnr\n"
            # )
        try:
            mp.set_start_method(method, force=True)
        except RuntimeError:
            # Already set (possibly by another library, or called too late)
            warnings.warn(
                f"GWSNR: Could not set multiprocessing start method to '{method}'. "
                "This is usually because the start method was already set by another library or your script, "
                "or because it is being set too late.\n"
                "If you need to control the multiprocessing start method, set it at the very beginning of your script.\n"
                "To use the 'spawn' method instead, set the environment variable GWSNR_USE_SPAWN=True before running your script.\n"
                "Command line (single line):\n"
                "    GWSNR_USE_SPAWN=True python yourscript.py\n"
                "In a Python script (before importing GWSNR):\n"
                "    import os\n"
                "    os.environ['GWSNR_USE_SPAWN'] = 'True'\n"
                "    import gwsnr\n"
            )
    else:
        # For Windows, default is already 'spawn', nothing to do.
        pass

# Call the function on package import
set_multiprocessing_start_method()


import warnings
import logging
from . import core
from .core import GWSNR
from ._version import __version__

# Package metadata
__all__ = ['GWSNR', 'core']
__author__ = 'Hemantakumar Phurailatpam <hemantaphurailatpam@gmail.com>'
__license__ = "MIT"
__email__ = "hemantaphurailatpam@gmail.com"
__maintainer__ = "Hemantakumar Phurailatpam"
__status__ = "Development"
__url__ = "https://github.com/hemantaph/gwsnr"
__description__ = "A Python package for calculating gravitational wave signal-to-noise ratios"
__version_info__ = tuple(map(int, __version__.split('.')))

warnings.filterwarnings("ignore", "Wswiglal-redir-stdio")
logging.getLogger(__name__).addHandler(logging.NullHandler())





# -*- coding: utf-8 -*-

import jax
# jax.config.update("jax_enable_x64", True)
import numpy as np
import jax.numpy as jnp
from jax import jit, vmap, lax
from ..numba import antenna_response_array

@jit
def find_index_1d_jax(x_array, x_new):
    """
    Find the index for cubic spline interpolation in 1D.
    Returns the index and a condition for edge handling.

    Parameters
    ----------
    x_array : jnp.ndarray
        The array of x values for interpolation. Must be sorted in ascending order.
    x_new : float or jnp.ndarray
        The new x value(s) to find the index for.

    Returns
    -------
    i : jnp.ndarray
        The index in `x_array` where `x_new` would fit, clipped to range [1, N-3] 
        where N is the length of x_array.
    condition_i : jnp.ndarray
        A condition indicating which interpolation branch to use:
        - 1: Use linear interpolation at the left edge (x_new <= x_array[1]).
        - 2: Use cubic interpolation in the middle.
        - 3: Use linear interpolation at the right edge (x_new >= x_array[N-2]).

    Notes
    -----
    Uses binary search with clipped indices to ensure valid 4-point stencils.
    The condition parameter determines linear vs cubic interpolation at boundaries.
    """
    N = x_array.shape[0]
    # Use jnp.searchsorted for robust, fast binary search.
    i = jnp.searchsorted(x_array, x_new, side='right') - 1
    i = jnp.clip(i, 1, N - 3)

    # Condition for edge handling (same as Numba version)
    condition_i = jnp.where(x_new <= x_array[1], 1,
                  jnp.where(x_new >= x_array[N-2], 3, 2))
    
    return i, condition_i

@jit
def spline_interp_4pts_jax(x_eval, x_pts, y_pts, condition_i):
    """
    Performs piecewise interpolation using 4 points with JAX compatibility.
    This function implements a piecewise interpolation scheme that uses:
    - Linear interpolation at the left boundary (condition_i=1)
    - Cubic interpolation in the middle region (condition_i=2)
    - Linear interpolation at the right boundary (condition_i=3)
    The cubic interpolation uses cubic Hermite spline coefficients for smooth
    interpolation between the middle two points, while the boundary regions
    use linear interpolation for stability.
    Parameters
    ----------
    x_eval : array_like
        The x-coordinate(s) where interpolation is to be evaluated.
    x_pts : array_like
        Array of 4 x-coordinates of the interpolation points, ordered as
        [x0, x1, x2, x3] where x1 and x2 are the main interpolation interval.
    y_pts : array_like
        Array of 4 y-coordinates corresponding to x_pts, ordered as
        [y0, y1, y2, y3].
    condition_i : int
        Interpolation mode selector:
        - 1: Linear interpolation using points (x0, y0) and (x1, y1)
        - 2: Cubic interpolation using all 4 points with x_eval in [x1, x2]
        - 3: Linear interpolation using points (x2, y2) and (x3, y3)

    Returns
    -------
    array_like
        Interpolated value(s) at x_eval using the specified interpolation method.

    Notes
    -----
    - The function handles degenerate cases where denominators are zero by
      returning appropriate fallback values (y0, y1, or y2 respectively).
    - Uses JAX's lax.switch for efficient conditional execution.
    - The cubic interpolation uses normalized parameter t = (x_eval - x1) / (x2 - x1).
    - Cubic coefficients follow the pattern: a*t³ + b*t² + c*t + d where:
    """

    # Linear at boundaries
    def left_branch(operands):
        x_eval, x_pts, y_pts = operands
        denom = x_pts[1] - x_pts[0]
        return y_pts[0] + (y_pts[1] - y_pts[0]) * (x_eval - x_pts[0]) / denom
    def right_branch(operands):
        x_eval, x_pts, y_pts = operands
        denom = x_pts[3] - x_pts[2]
        return y_pts[2] + (y_pts[3] - y_pts[2]) * (x_eval - x_pts[2]) / denom
    # Cubic interpolation in the middle
    def cubic_branch(operands):
        x_eval, x_pts, y_pts = operands
        x0, x1, x2, x3 = x_pts[0], x_pts[1], x_pts[2], x_pts[3]
        y0, y1, y2, y3 = y_pts[0], y_pts[1], y_pts[2], y_pts[3]
        denom = x2 - x1
        t = (x_eval - x1) / denom

        # --- Cubic Hermite spline tangents ---
        m1 = ((y2 - y1) / (x2 - x1)) * ((x1 - x0) / (x2 - x0)) + ((y1 - y0) / (x1 - x0)) * ((x2 - x1) / (x2 - x0))
        m2 = ((y3 - y2) / (x3 - x2)) * ((x2 - x1) / (x3 - x1)) + ((y2 - y1) / (x2 - x1)) * ((x3 - x2) / (x3 - x1))

        # --- Hermite cubic basis ---
        h00 = (2 * t**3 - 3 * t**2 + 1)
        h10 = (t**3 - 2 * t**2 + t)
        h01 = (-2 * t**3 + 3 * t**2)
        h11 = (t**3 - t**2)

        # Interpolated value
        return h00 * y1 + h10 * m1 * denom + h01 * y2 + h11 * m2 * denom
    
    return jax.lax.switch(
        condition_i - 1,  # 1 -> 0, 2 -> 1, 3 -> 2
        [left_branch, cubic_branch, right_branch],
        (x_eval, x_pts, y_pts)
    )

######### Interpolation 4D #########
@jit
def spline_interp_4x4x4x4pts_jax(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new, mtot_new, a1_new, a2_new):
    """
    Function that performs the FULL 4D interpolation for a SINGLE point.
    This function finds indices, slices data, and then uses vmap internally
    to perform interpolation efficiently without Python loops.
    """
    # 1. Find indices and conditions for the new point
    q_idx, int_q = find_index_1d_jax(q_array, q_new)
    m_idx, int_m = find_index_1d_jax(mtot_array, mtot_new)
    a1_idx, int_a1 = find_index_1d_jax(a1_array, a1_new)
    a2_idx, int_a2 = find_index_1d_jax(a2_array, a2_new)

    # 2. Slice the small 4x4x4x4 cube of data and the corresponding 4-point coordinate arrays
    q_pts = lax.dynamic_slice(q_array, (q_idx - 1,), (4,))
    m_pts = lax.dynamic_slice(mtot_array, (m_idx - 1,), (4,))
    a1_pts = lax.dynamic_slice(a1_array, (a1_idx - 1,), (4,))
    a2_pts = lax.dynamic_slice(a2_array, (a2_idx - 1,), (4,))
    data_cube = lax.dynamic_slice(snrpartialscaled_array, (q_idx-1, m_idx-1, a1_idx-1, a2_idx-1), (4, 4, 4, 4))

    # 3. Perform 4D interpolation using vmap to eliminate loops
    # Interpolate along the last dimension (a2) for each of the 4x4x4 slices
    interp_along_a2 = vmap(
        lambda y_slice: spline_interp_4pts_jax(a2_new, a2_pts, y_slice, int_a2)
    )(data_cube.reshape(64, 4)).reshape(4, 4, 4)

    # Interpolate along the next dimension (a1)
    interp_along_a1 = vmap(
        lambda y_slice: spline_interp_4pts_jax(a1_new, a1_pts, y_slice, int_a1)
    )(interp_along_a2.reshape(16, 4)).reshape(4, 4)

    # Interpolate along the mtot dimension
    interp_along_m = vmap(
        lambda y_slice: spline_interp_4pts_jax(mtot_new, m_pts, y_slice, int_m)
    )(interp_along_a1)

    # Final interpolation along the q dimension
    final_snr = spline_interp_4pts_jax(q_new, q_pts, interp_along_m, int_q)

    return final_snr

@jit
def spline_interp_4x4x4x4pts_batched_jax(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch):
    """
    Perform batched 4D cubic spline interpolation using JAX vectorization.
    """
    # Vectorize the complete single-point interpolation function.
    # This is the only vmap call needed at the top level.
    vmapped_interpolator = vmap(
        spline_interp_4x4x4x4pts_jax,
        in_axes=(None, None, None, None, None, 0, 0, 0, 0)
    )

    # Call the fully vectorized function on the batch data.
    return vmapped_interpolator(
        q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array,
        q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch
    )

def get_interpolated_snr_aligned_spins_jax(mass_1, mass_2, luminosity_distance, theta_jn, psi, geocent_time, ra, dec, a_1, a_2, detector_tensor, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, batch_size=100000):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """

    # get array of antenna response
    Fp, Fc = antenna_response_array(ra, dec, geocent_time, psi, detector_tensor)

    snr, snr_effective, snr_partial_, d_eff = get_interpolated_snr_aligned_spins_helper(
        jnp.array(mass_1), 
        jnp.array(mass_2), 
        jnp.array(luminosity_distance), 
        jnp.array(theta_jn), 
        jnp.array(a_1), 
        jnp.array(a_2), 
        jnp.array(snr_partialscaled), 
        jnp.array(ratio_arr), 
        jnp.array(mtot_arr), 
        jnp.array(a1_arr), 
        jnp.array(a_2_arr),
        jnp.array(Fp),
        jnp.array(Fc),
        jnp.array(detector_tensor)
    )

    return np.array(snr), np.array(snr_effective), np.array(snr_partial_), np.array(d_eff)

@jit
def get_interpolated_snr_aligned_spins_helper(mass_1, mass_2, luminosity_distance, theta_jn, a_1, a_2, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, Fp, Fc, detector_tensor):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """
    
    det_len = detector_tensor.shape[0]
    size = mass_1.shape[0]
    mtot = mass_1 + mass_2
    ratio = mass_2 / mass_1

    Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
    A1 = Mc ** (5.0 / 6.0)
    ci_2 = jnp.cos(theta_jn) ** 2
    ci_param = ((1 + jnp.cos(theta_jn) ** 2) / 2) ** 2

    snr_partial_ = jnp.zeros((det_len,size))
    d_eff = jnp.zeros((det_len,size))
    snr = jnp.zeros((det_len,size))
    # loop over the detectors
    for j in range(det_len):
        snr_partial_ = snr_partial_.at[j].set(
            jnp.array(spline_interp_4x4x4x4pts_batched_jax(
                q_array=ratio_arr,
                mtot_array=mtot_arr,
                a1_array=a1_arr,
                a2_array=a_2_arr,
                snrpartialscaled_array=snr_partialscaled[j],
                q_new_batch=ratio,
                mtot_new_batch=mtot,
                a1_new_batch=a_1,
                a2_new_batch=a_2,
            ))
        )

        # calculate the effective distance
        d_eff = d_eff.at[j].set(
            luminosity_distance / jnp.sqrt(Fp[j] ** 2 * ci_param + Fc[j] ** 2 * ci_2)
        )
        # snr[j] = snr_partial_buffer * A1 / d_eff[j]
        snr = snr.at[j].set(
            snr_partial_[j] * A1 / d_eff[j]
        )
    
    snr_effective = jnp.sqrt(jnp.sum(snr ** 2, axis=0))

    return snr, snr_effective, snr_partial_, d_eff

######### Interpolation 2D #########
@jit
def spline_interp_4x4pts_jax(q_array, mtot_array, snrpartialscaled_array, q_new, mtot_new):
    """
    Function that performs the FULL 2D interpolation for a SINGLE point.
    This function finds indices, slices data, and then uses vmap internally
    to perform interpolation efficiently without Python loops.
    """

    # 1. Find indices and conditions for the new point
    q_idx, int_q = find_index_1d_jax(q_array, q_new)
    m_idx, int_m = find_index_1d_jax(mtot_array, mtot_new)

    # 2. Slice the small 4x4x4x4 cube of data and the corresponding 4-point coordinate arrays
    q_pts = lax.dynamic_slice(q_array, (q_idx - 1,), (4,))
    m_pts = lax.dynamic_slice(mtot_array, (m_idx - 1,), (4,))
    F = lax.dynamic_slice(snrpartialscaled_array, (q_idx - 1, m_idx - 1), (4, 4))

    # Interpolate along the mtot dimension
    interp_along_m = vmap(
        lambda y_slice: spline_interp_4pts_jax(mtot_new, m_pts, y_slice, int_m)
    )(F)

    # Now, interpolate the resulting 4 values along the q axis
    final_snr = spline_interp_4pts_jax(q_new, q_pts, interp_along_m, int_q)

    return final_snr

# Vectorized version
@jit
def spline_interp_4x4pts_batched_jax(q_array, mtot_array, snrpartialscaled_array, q_new_batch, mtot_new_batch):
    """
    Perform batched 2D cubic spline interpolation using JAX vectorization.
    """

    # Vectorize the complete single-point interpolation function.
    # This is the only vmap call needed at the top level.
    vmapped_interpolator = vmap(
        spline_interp_4x4pts_jax,
        in_axes=(None, None, None, 0, 0)
    )

    # Call the fully vectorized function on the batch data.
    return vmapped_interpolator(
        q_array, mtot_array, snrpartialscaled_array, q_new_batch, mtot_new_batch
    )

def get_interpolated_snr_no_spins_jax(mass_1, mass_2, luminosity_distance, theta_jn, psi, geocent_time, ra, dec, a_1, a_2, detector_tensor, snr_partialscaled, ratio_arr, mtot_arr, a1_arr, a_2_arr, batch_size=100000):
    """
    Calculate interpolated signal-to-noise ratio (SNR) for aligned spin gravitational wave signals using JAX.
    This function computes the SNR for gravitational wave signals with aligned spins across multiple 
    detectors using 4D cubic spline interpolation. It calculates the effective distance, partial SNR,
    and combines results from multiple detectors to produce the effective SNR.

    Parameters
    ----------
    mass_1 : jax.numpy.ndarray
        Primary mass of the binary system in solar masses.
    mass_2 : jax.numpy.ndarray  
        Secondary mass of the binary system in solar masses.
    luminosity_distance : jax.numpy.ndarray
        Luminosity distance to the source in Mpc.
    theta_jn : jax.numpy.ndarray
        Inclination angle between the orbital angular momentum and line of sight in radians.
    psi : jax.numpy.ndarray
        Polarization angle in radians.
    geocent_time : jax.numpy.ndarray
        GPS time of coalescence at the geocenter in seconds.
    ra : jax.numpy.ndarray
        Right ascension of the source in radians.
    dec : jax.numpy.ndarray
        Declination of the source in radians.
    a_1 : jax.numpy.ndarray
        Dimensionless spin magnitude of the primary black hole.
    a_2 : jax.numpy.ndarray
        Dimensionless spin magnitude of the secondary black hole.
    detector_tensor : jax.numpy.ndarray
        Detector tensor array containing detector response information.
        Shape: (n_detectors, ...)
    snr_partialscaled : jax.numpy.ndarray
        Pre-computed scaled partial SNR values for interpolation.
        Shape: (n_detectors, ...)
    ratio_arr : jax.numpy.ndarray
        Mass ratio grid points for interpolation (q = m2/m1).
    mtot_arr : jax.numpy.ndarray
        Total mass grid points for interpolation.
    a1_arr : jax.numpy.ndarray
        Primary spin grid points for interpolation.
    a_2_arr : jax.numpy.ndarray
        Secondary spin grid points for interpolation.

    Returns
    -------
    snr : jax.numpy.ndarray
        SNR values for each detector. Shape: (n_detectors, n_samples)
    snr_effective : jax.numpy.ndarray
        Effective SNR combining all detectors. Shape: (n_samples,)
    snr_partial_ : jax.numpy.ndarray
        Interpolated partial SNR values for each detector. Shape: (n_detectors, n_samples)
    d_eff : jax.numpy.ndarray
        Effective distance for each detector accounting for antenna response.
        Shape: (n_detectors, n_samples)

    Notes
    -----
    - Uses 4D cubic spline interpolation for efficient SNR calculation
    - Assumes aligned spins (no precession)
    - Effective SNR is calculated as sqrt(sum(SNR_i^2)) across detectors
    - Chirp mass and inclination-dependent factors are computed analytically
    """

    # get array of antenna response
    Fp, Fc = antenna_response_array(ra, dec, geocent_time, psi, detector_tensor)

    snr, snr_effective, snr_partial_, d_eff = get_interpolated_snr_no_spins_helper(
        jnp.array(mass_1), 
        jnp.array(mass_2), 
        jnp.array(luminosity_distance), 
        jnp.array(theta_jn), 
        jnp.array(snr_partialscaled), 
        jnp.array(ratio_arr), 
        jnp.array(mtot_arr), 
        jnp.array(Fp),
        jnp.array(Fc),
        detector_tensor
    )

    return np.array(snr), np.array(snr_effective), np.array(snr_partial_), np.array(d_eff)

@jit
def get_interpolated_snr_no_spins_helper(mass_1, mass_2, luminosity_distance, theta_jn, snr_partialscaled, ratio_arr, mtot_arr, Fp, Fc, detector_tensor):
    """
    Function to calculate the interpolated snr for a given set of parameters
    """

    size = mass_1.shape[0]
    det_len = detector_tensor.shape[0]
    mtot = mass_1 + mass_2
    ratio = mass_2 / mass_1

    Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
    A1 = Mc ** (5.0 / 6.0)
    ci_2 = jnp.cos(theta_jn) ** 2
    ci_param = ((1 + jnp.cos(theta_jn) ** 2) / 2) ** 2

    snr_partial_ = jnp.zeros((det_len,size))
    d_eff = jnp.zeros((det_len,size))
    snr = jnp.zeros((det_len,size))
    # loop over the detectors
    for j in range(det_len):
        snr_partial_ = snr_partial_.at[j].set(
            jnp.array(spline_interp_4x4pts_batched_jax(
                q_array=ratio_arr,
                mtot_array=mtot_arr,
                snrpartialscaled_array=snr_partialscaled[j],
                q_new_batch=ratio,
                mtot_new_batch=mtot,
            ))
        )

        d_eff = d_eff.at[j].set(
            luminosity_distance / jnp.sqrt(Fp[j] ** 2 * ci_param + Fc[j] ** 2 * ci_2)
        )
        # snr[j] = snr_partial_buffer * A1 / d_eff[j]
        snr = snr.at[j].set(
            snr_partial_[j] * A1 / d_eff[j]
        )
    
    snr_effective = jnp.sqrt(jnp.sum(snr ** 2, axis=0))

    return snr, snr_effective, snr_partial_, d_eff

# ########### For testing only ##################
# def cubic_hermite_spline_spline(p, t):
    
#     M = 0.5 * jnp.array([
#         [0,  2,  0,  0],
#         [-1, 0,  1,  0],
#         [2, -5,  4, -1],
#         [-1, 3, -3,  1]
#     ])
#     T = jnp.array([1.0, t, t**2, t**3])
#     return T @ M @ p

# @jit
# def find_index_1d(x_array, x_new):
#     N = x_array.shape[0]
#     i = jnp.sum(x_array <= x_new) - 1
#     return jnp.clip(i, 1, N - 3)

# @jit
# def cubic_hermite_spline_4d_interp_single(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new, mtot_new, a1_new, a2_new):
#     q_idx = find_index_1d(q_array, q_new)
#     m_idx = find_index_1d(mtot_array, mtot_new)
#     a1_idx = find_index_1d(a1_array, a1_new)
#     a2_idx = find_index_1d(a2_array, a2_new)

#     # Dynamic slices (for 4x4 neighborhood)
#     F = lax.dynamic_slice(snrpartialscaled_array, (q_idx-1, m_idx-1, a1_idx-1, a2_idx-1), (4,4,4,4))
#     qs = lax.dynamic_slice(q_array, (q_idx - 1,), (4,))
#     ms = lax.dynamic_slice(mtot_array, (m_idx - 1,), (4,))
#     a1s = lax.dynamic_slice(a1_array, (a1_idx - 1,), (4,))
#     a2s = lax.dynamic_slice(a2_array, (a2_idx - 1,), (4,))

#     # Relative coordinates
#     tq = (q_new - qs[1]) / (qs[2] - qs[1])
#     tm = (mtot_new - ms[1]) / (ms[2] - ms[1])
#     ta1 = (a1_new - a1s[1]) / (a1s[2] - a1s[1])
#     ta2 = (a2_new - a2s[1]) / (a2s[2] - a2s[1])

#     # Tricubic interpolation logic, extended to 4D:
#     temp_q = jnp.zeros(4)
#     for i in range(4):
#         temp_m = jnp.zeros(4)
#         for j in range(4):
#             temp_a1 = jnp.zeros(4)
#             for k in range(4):
#                 # Interpolate along a2 (last axis)
#                 temp_a1= temp_a1.at[k].set(cubic_hermite_spline(F[i, j, k, :], ta2))
#             # Interpolate along a1
#             temp_m = temp_m.at[j].set(cubic_hermite_spline(temp_a1, ta1))
#         # Interpolate along mtot
#         temp_q = temp_q.at[i].set(cubic_hermite_spline(temp_m, tm))
#     # Interpolate along q
#     snr_new = cubic_hermite_spline(temp_q, tq)
#     return snr_new

# @jit
# def batched_cubic_hermite_spline_4d(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch):
#     # Vectorize only over q_new and mtot_new
#     vmapped_interp = vmap(
#         lambda q, m, a1, a2: cubic_hermite_spline_4d_interp_single(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q, m, a1, a2),
#         in_axes=(0, 0, 0, 0)
#     )
#     return vmapped_interp(q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch)

# @jit
# def cubic_spline_4d_jax(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new, mtot_new, a1_new, a2_new, int_q, int_m, int_a1, int_a2):
#     """
#     Perform 4D cubic spline interpolation using JAX operations.
#     This function interpolates a 4D array (snrpartialscaled_array) at specified points
#     using cubic spline interpolation. The interpolation is performed sequentially
#     along each dimension: first a2, then a1, then mtot, and finally q.

#     Parameters
#     ----------
#     q_array : jax.numpy.ndarray
#         1D array containing the q-dimension coordinate values.
#     mtot_array : jax.numpy.ndarray
#         1D array containing the total mass dimension coordinate values.
#     a1_array : jax.numpy.ndarray
#         1D array containing the first spin parameter dimension coordinate values.
#     a2_array : jax.numpy.ndarray
#         1D array containing the second spin parameter dimension coordinate values.
#     snrpartialscaled_array : jax.numpy.ndarray
#         4D array containing the SNR partial scaled values to be interpolated.
#         Shape should be (len(q_array), len(mtot_array), len(a1_array), len(a2_array)).
#     q_new : float
#         New q value at which to interpolate.
#     mtot_new : float
#         New total mass value at which to interpolate.
#     a1_new : float
#         New first spin parameter value at which to interpolate.
#     a2_new : float
#         New second spin parameter value at which to interpolate.
#     int_q : int
#         edge condition for q interpolation. Refer to `find_index_1d_numba` for details.
#     int_m : int
#         edge condition for mtot interpolation. Refer to `find_index_1d_numba` for details.
#     int_a1 : int
#         edge condition for a1 interpolation. Refer to `find_index_1d_numba` for details.
#     int_a2 : int
#         edge condition for a2 interpolation. Refer to `find_index_1d_numba` for details.

#     Returns
#     -------
#     float
#         Interpolated SNR value at the specified (q_new, mtot_new, a1_new, a2_new) point.

#     Notes
#     -----
#     This function uses nested loops to perform interpolation sequentially along each
#     dimension. It relies on helper functions `find_index_1d_jax` for finding array
#     indices and `cubic_function_4pts_jax` for 1D cubic interpolation using 4 points.
#     The interpolation process:
#     1. Find indices and interpolation weights for each dimension
#     2. Interpolate along a2 dimension for each combination of q, mtot, a1 indices
#     3. Interpolate along a1 dimension using results from step 2
#     4. Interpolate along mtot dimension using results from step 3
#     5. Interpolate along q dimension to get the final result
#     """

#     # Find indices and conditions for q, mtot, a1, a2
#     partialsnr_along_q = jnp.zeros(4)
#     for i in range(4):
#         partialsnr_along_m = jnp.zeros(4)
#         for j in range(4):
#             partialsnr_along_a1 = jnp.zeros(4)
#             for k in range(4):
#                 # Interpolate along a2
#                 partialsnr_along_a1 = partialsnr_along_a1.at[k].set(
#                     cubic_function_4pts_jax(
#                         x_eval=a2_new, 
#                         x_pts=a2_array, 
#                         y_pts=snrpartialscaled_array[i, j, k, :], 
#                         condition_i=int_a2
#                     )
#                 )
#             # Interpolate along a1
#             partialsnr_along_m = partialsnr_along_m.at[j].set(
#                 cubic_function_4pts_jax(a1_new, a1_array, partialsnr_along_a1, int_a1)
#             )
#         # Interpolate along mtot
#         partialsnr_along_q = partialsnr_along_q.at[i].set(
#             cubic_function_4pts_jax(mtot_new, mtot_array, partialsnr_along_m, int_m)
#         )
#     # Interpolate along q
#     final_snr = cubic_function_4pts_jax(q_new, q_array, partialsnr_along_q, int_q)
    
#     return final_snr

# def _helper(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new, mtot_new, a1_new, a2_new):
#     """
#     Processes a single set of parameters. This function is designed to be vectorized by vmap.
#     All array arguments are passed explicitly.
#     """
#     # Find the index for each parameter for a single item
#     q_idx, q_condition_i = find_index_1d_jax(q_array, q_new)
#     m_idx, m_condition_i = find_index_1d_jax(mtot_array, mtot_new)
#     a1_idx, a1_condition_i = find_index_1d_jax(a1_array, a1_new)
#     a2_idx, a2_condition_i = find_index_1d_jax(a2_array, a2_new)

#     # Slice the 1D arrays to get the 4 points around the index
#     q_array_slice = lax.dynamic_slice(q_array, (q_idx - 1,), (4,))
#     m_array_slice = lax.dynamic_slice(mtot_array, (m_idx - 1,), (4,))
#     a1_array_slice = lax.dynamic_slice(a1_array, (a1_idx - 1,), (4,))
#     a2_array_slice = lax.dynamic_slice(a2_array, (a2_idx - 1,), (4,))

#     # Slice the main 4D data cube
#     cube_4x4x4x4 = lax.dynamic_slice(
#         snrpartialscaled_array,
#         (q_idx - 1, m_idx - 1, a1_idx - 1, a2_idx - 1),
#         (4, 4, 4, 4)
#     )

#     # Return all results for the single item
#     return (q_idx, q_condition_i, m_idx, m_condition_i, a1_idx, a1_condition_i, a2_idx, a2_condition_i,
#             q_array_slice, m_array_slice, a1_array_slice, a2_array_slice, cube_4x4x4x4)

# # The main function is JIT-compiled once.
# @jit
# def cubic_spline_4d_batched_jax(q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array, q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch):
#     """
#     Perform batched 4D cubic spline interpolation using JAX vectorization.
#     """
#     # Vectorize the helper function directly.
#     # Use in_axes to specify that the first 5 arguments are static (None)
#     # and the next 4 are batched over the first axis (0).
#     vmapped_helper = vmap(
#         _helper,
#         in_axes=(None, None, None, None, None, 0, 0, 0, 0)
#     )

#     # Call the vectorized function.
#     # The static arrays are passed once and are broadcasted by vmap, not iterated over.
#     q_idx, q_condition_i, m_idx, m_condition_i, a1_idx, a1_condition_i, a2_idx, a2_condition_i, q_array_slice, m_array_slice, a1_array_slice, a2_array_slice, cube_4x4x4x4 = vmapped_helper(
#         q_array, mtot_array, a1_array, a2_array, snrpartialscaled_array,
#         q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch
#     )

#     results = vmap(cubic_spline_4d_jax)(
#         q_array_slice, m_array_slice, a1_array_slice, a2_array_slice, cube_4x4x4x4,
#         q_new_batch, mtot_new_batch, a1_new_batch, a2_new_batch
#     )

#     return results


# #################


# -*- coding: utf-8 -*-
"""
This module contains the ANNModelGenerator class which is used to generate the ANN (Artificial Neural Network) model that can be used to predict the SNR of the GW events.
"""

import numpy as np
import os
import pickle
from ..core import GWSNR
from ..utils import append_json, get_param_from_json, load_json, load_ann_h5
from scipy.optimize import curve_fit
import jax
# jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp

import tensorflow as tf
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, accuracy_score

class ANNModelGenerator():
    """
    ANNModelGenerator class is used to generate the ANN model that can be used to predict the SNR of the GW events.

    Parameters
    ----------
    npool : `int`
        Number of processors to use for parallel processing.
        Default is 4.
    gwsnr_verbose : `bool`
        If True, print the progress of the GWSNR calculation.
        Default is True.
    snr_th : `float`
        SNR threshold for the error calculation.
        Default is 8.0.
    waveform_approximant : `str`
        Waveform approximant to be used for the GWSNR calculation and the ANN model.
        Default is "IMRPhenomXPHM".
    **kwargs : `dict`
        Keyword arguments for the GWSNR class.
        To see the list of available arguments, 
        >>> from gwsnr import GWSNR
        >>> help(GWSNR)
        
    Examples
    --------
    >>> from gwsnr import ANNModelGenerator
    >>> amg = ANNModelGenerator()
    >>> amg.ann_model_training(gw_param_dict='gw_param_dict.json') # training the ANN model with pre-generated parameter points

    """
    
    def __init__(self,
        directory="./gwsnr_data",
        npool=4,
        gwsnr_verbose=True,
        snr_th=8.0,
        snr_method="interpolation_aligned_spins",
        waveform_approximant="IMRPhenomXPHM",
        **kwargs,  # ler and gwsnr arguments
    ):

        self.directory = directory
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)

        self.ann_model = self.ann_model_4layers
        self.ann = None
        self.scaler = None

        self.gwsnr_args = dict(
            npool=npool,
            # gwsnr args
            mtot_min=2*4.98, # 4.98 Mo is the minimum component mass of BBH systems in GWTC-3
            mtot_max=2*112.5+10.0, # 112.5 Mo is the maximum component mass of BBH systems in GWTC-3. 10.0 Mo is added to avoid edge effects.
            ratio_min=0.1,
            ratio_max=1.0,
            spin_max=0.99,
            mtot_resolution=200,
            ratio_resolution=20,
            spin_resolution=10,
            sampling_frequency=2048.0,
            waveform_approximant=waveform_approximant,
            minimum_frequency=20.0,
            snr_method="interpolation_aligned_spins",
            psds=None,
            ifos=None,
            interpolator_dir="./interpolator_pickle",
            create_new_interpolator=False,
            gwsnr_verbose=gwsnr_verbose,
            multiprocessing_verbose=True,
            mtot_cut=False,
            snr_th=snr_th,
        )
        self.gwsnr_args.update(kwargs)

        # gwsnr initialization
        # spinless
        self.gwsnr = GWSNR(
            npool=self.gwsnr_args['npool'],
            # gwsnr args
            mtot_min=self.gwsnr_args['mtot_min'],
            mtot_max=self.gwsnr_args['mtot_max'],
            ratio_min=self.gwsnr_args['ratio_min'],
            ratio_max=self.gwsnr_args['ratio_max'],
            spin_max=self.gwsnr_args['spin_max'],
            mtot_resolution=self.gwsnr_args['mtot_resolution'],
            ratio_resolution=self.gwsnr_args['ratio_resolution'],
            spin_resolution=self.gwsnr_args['spin_resolution'],
            sampling_frequency=self.gwsnr_args['sampling_frequency'],
            waveform_approximant=self.gwsnr_args['waveform_approximant'],
            minimum_frequency=self.gwsnr_args['minimum_frequency'],
            snr_method=snr_method,
            psds=self.gwsnr_args['psds'],
            ifos=self.gwsnr_args['ifos'],
            interpolator_dir=self.gwsnr_args['interpolator_dir'],
            create_new_interpolator=self.gwsnr_args['create_new_interpolator'],
            gwsnr_verbose=self.gwsnr_args['gwsnr_verbose'],
            multiprocessing_verbose=self.gwsnr_args['multiprocessing_verbose'],
            mtot_cut=self.gwsnr_args['mtot_cut'],
            snr_th=self.gwsnr_args['snr_th'],
        )

    def get_input_output_data(self, params=None, randomize=True):

        params = self.get_parameters(params)

        if randomize:
            idx = np.random.choice(len(params['mass_1']), size=len(params['mass_1']), replace=False)
        else:
            idx = np.arange(len(params['mass_1']))

        for key, value in params.items():
            params[key] = np.array(value)[idx]

        X1 = self.get_input_data(params=params)
        # output data
        # get snr for y train
        det_ = self.gwsnr.detector_list[0]
        if 'snr_net' in params:
            y1 = np.array(params['snr_net'])
        elif det_ in params:
            y1 = np.array(params[det_])
        else:
            raise ValueError("Output data (snr) not found")

        return X1, y1
        
    def get_input_data(self, params):
        """
        Function to generate input and output data for the neural network

        Parameters:
        idx: index of the parameter points
        params: dictionary of parameter points
            params.keys() = ['mass_1', 'mass_2', 'luminosity_distance', 'theta_jn', 'psi', 'geocent_time', 'ra', 'dec', 'a_1', 'a_2', 'tilt_1', 'tilt_2', 'L1']

        Returns:
        X: input data, [snr_partial_[0], amp0[0], eta, chi_eff, theta_jn]
        y: output data, [L1]
        """

        mass_1 = np.array(params['mass_1'])
        mass_2 = np.array(params['mass_2'])
        luminosity_distance = np.array(params['luminosity_distance'])
        theta_jn = np.array(params['theta_jn'])
        psi = np.array(params['psi'])
        geocent_time = np.array(params['geocent_time'])
        ra = np.array(params['ra'])
        dec = np.array(params['dec'])
        a_1 = np.array(params['a_1'])
        a_2 = np.array(params['a_2'])
        tilt_1 = np.array(params['tilt_1'])
        tilt_2 = np.array(params['tilt_2'])
        # effective spin
        chi_eff = (mass_1 * a_1 * np.cos(tilt_1) + mass_2 * a_2 * np.cos(tilt_2)) / (mass_1 + mass_2)

        # to get the components of the spin aligned with angular momentum
        a_1 = a_1 * np.cos(tilt_1)
        a_2 = a_2 * np.cos(tilt_2)
        
        detector_tensor = np.array(self.gwsnr.detector_tensor_list)
        len_ = len(detector_tensor)
        if len_ != 1:
            raise ValueError("Only one detector is allowed")
        
        # calculate the snr_partial_ using natural cubic spline interpolation
        _, _, snr_partial_, d_eff = self.gwsnr.get_interpolated_snr(
            mass_1 = np.array(mass_1),
            mass_2 = np.array(mass_2),
            luminosity_distance = np.array(luminosity_distance),
            theta_jn = np.array(theta_jn),
            psi = np.array(psi),
            geocent_time = np.array(geocent_time),
            ra = np.array(ra),
            dec = np.array(dec),
            a_1 = np.array(a_1),
            a_2 = np.array(a_2),
            detector_tensor = detector_tensor,
            snr_partialscaled = np.array(self.gwsnr.snr_partialsacaled_list),
            ratio_arr = np.array(self.gwsnr.ratio_arr),
            mtot_arr = np.array(self.gwsnr.mtot_arr),
            a1_arr = np.array(self.gwsnr.a_1_arr),
            a_2_arr = np.array(self.gwsnr.a_2_arr),
        )

        # calculate the effective amplitude
        Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
        eta = mass_1 * mass_2/(mass_1 + mass_2)**2.
        # A1 = Mc ** (5.0 / 6.0)
        amp0 =  Mc ** (5.0 / 6.0) / np.array(d_eff)[0]

        # input data
        X1 = np.vstack([np.array(snr_partial_)[0], amp0, eta, chi_eff, theta_jn]).T

        return X1

    def standard_scaling_initialization(self, X_train):

        sc = StandardScaler()
        X_train = sc.fit_transform(X_train)
        self.scaler = sc

        return X_train, sc

    def ann_model_4layers(self, 
        num_nodes_list,
        activation_fn_list,
        optimizer,
        loss,
        metrics,
    ):
        ann = tf.keras.models.Sequential()

        for i in range(1, len(num_nodes_list)):
            ann.add(tf.keras.layers.Dense(units=num_nodes_list[i], activation=activation_fn_list[i]))

        # compile the ANN
        ann.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        return ann

    def get_parameters(self, gw_param_dict):

        # get the parameters
        if isinstance(gw_param_dict, str):
            path_ = f"{gw_param_dict}"
            gw_param_dict = get_param_from_json(path_)
        elif isinstance(gw_param_dict, dict):
            pass
        else:
            raise ValueError("gw_param_dict must be a dictionary or a json file")

        return gw_param_dict

    def get_scaled_data(self, gw_param_dict, randomize=True, test_size=0.1, random_state=42):

        gw_param_dict = self.get_parameters(gw_param_dict)

        # input and output data
        X, y = self.get_input_output_data(params=gw_param_dict, randomize=randomize)

        # scaling
        X = self.scaler.transform(X)

        return X, y

    def ann_model_training(self,
        gw_param_dict,
        randomize=True,
        test_size=0.1,
        random_state=42,
        num_nodes_list = [5, 32, 32, 1],
        activation_fn_list = ['relu', 'relu', 'sigmoid', 'linear'],
        optimizer='adam',
        loss='mean_squared_error',
        metrics=['accuracy'],
        batch_size=32,
        epochs=100,
        error_adjustment_snr_range=[4,10],
        ann_file_name = 'ann_model.h5',
        scaler_file_name = 'scaler.pkl',
        error_adjustment_file_name='error_adjustment.json',
        ann_path_dict_file_name='ann_path_dict.json',
    ):

        # input and output data
        X, y = self.get_input_output_data(params=gw_param_dict, randomize=randomize)

        # split the data
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)

        # scaling
        X_scaled, scaler = self.standard_scaling_initialization(X_train)
        self.X_test = scaler.transform(X_test)
        self.y_test = y_test
        # save the scaler
        pickle.dump(scaler, open(f'{self.directory}/{scaler_file_name}', 'wb'))

        # initialize the ANN
        ann = self.ann_model(num_nodes_list, activation_fn_list, optimizer, loss, metrics)

        # fit the ANN to the training set
        ann.fit(X_scaled, y_train, batch_size=batch_size, epochs=epochs)
        # for testing
        self.ann = ann
        # save the model
        ann.save(f'{self.directory}/{ann_file_name}')

        # error adjustment
        y_pred = np.array(ann.predict(self.X_test)).flatten()
        self.error_adjustment = self.helper_error_adjustment(y_pred, y_test, snr_range=error_adjustment_snr_range)
        append_json(f'{self.directory}/{error_adjustment_file_name}', self.error_adjustment, replace=True)

        # print the results
        print(f"scaler saved at: {self.directory}/{scaler_file_name}")
        print(f"model saved at: {self.directory}/{ann_file_name}")
        print(f"error adjustment saved at: {self.directory}/{error_adjustment_file_name}")

        # save the path of the model, scaler, and error adjustment
        self.save_ann_path_dict(
            ann_file_name=ann_file_name, 
            scaler_file_name=scaler_file_name, 
            error_adjustment_file_name=error_adjustment_file_name,
            ann_path_dict_file_name=ann_path_dict_file_name,
        )

    def pdet_error(self, gw_param_dict=None, randomize=True, error_adjustment=True):

        snr_threshold = self.gwsnr_args['snr_th']   # snr threshold

        if gw_param_dict is not None:
            gw_param_dict = self.get_parameters(gw_param_dict)
            # input and output data
            X_test, y_test = self.get_input_output_data(params=gw_param_dict, randomize=randomize)
            X_test = self.scaler.transform(X_test)
        else:
            X_test = self.X_test
            y_test = self.y_test

        # calculate the error
        y_pred = self.ann.predict(X_test).flatten()

        # error adjustment
        if error_adjustment:
            adjustment_dict = self.error_adjustment
            a = adjustment_dict['slope']
            b = adjustment_dict['intercept']
            y_pred = y_pred-(a*y_pred + b)

        len1 = len(y_pred)
        len2 = np.sum((y_pred>snr_threshold) != (y_test>snr_threshold))
        error = len2/len1*100
        print(f"Error: {error:.2f}%")

        return error, y_pred, y_test

    def save_ann_path_dict(self, 
        ann_file_name='ann_model.h5', 
        scaler_file_name='scaler.pkl', error_adjustment_file_name='error_adjustment.json',
        ann_path_dict_file_name='ann_path_dict.json',
    ):

        if not os.path.exists(f'{self.directory}/{ann_path_dict_file_name}'):
            ann_path_dict = {}
        else:
            ann_path_dict = load_json(f'{self.directory}/{ann_path_dict_file_name}')

        # check if the files exist
        if not os.path.exists(f'{self.directory}/{ann_file_name}'):
            raise ValueError("Model file does not exist")
        if not os.path.exists(f'{self.directory}/{scaler_file_name}'):
            raise ValueError("Scaler file does not exist")
        if not os.path.exists(f'{self.directory}/{error_adjustment_file_name}'):
            raise ValueError("Error adjustment file does not exist")
            
        ann_path_dict_ = {
            self.gwsnr.detector_list[0]: {
                "model_path": f'{self.directory}/{ann_file_name}',
                "scaler_path": f'{self.directory}/{scaler_file_name}',
                "error_adjustment_path": f'{self.directory}/{error_adjustment_file_name}',
                "sampling_frequency": self.gwsnr_args['sampling_frequency'],
                "minimum_frequency": self.gwsnr_args['minimum_frequency'],
                "waveform_approximant": self.gwsnr_args['waveform_approximant'], 
                "snr_th": self.gwsnr_args['snr_th'],},
        }

        ann_path_dict.update(ann_path_dict_)
        append_json(f'{self.directory}/{ann_path_dict_file_name}', ann_path_dict, replace=True)
        print(f"ann path dict saved at: {self.directory}/{ann_path_dict_file_name}")

        return ann_path_dict

    def pdet_confusion_matrix(self, gw_param_dict=None, randomize=True, snr_threshold=8.0):

        if gw_param_dict is not None:
            gw_param_dict = self.get_parameters(gw_param_dict)
            # input and output data
            X_test, y_test = self.get_input_output_data(params=gw_param_dict, randomize=randomize)
            X_test = self.scaler.transform(X_test)
        else:
            X_test = self.X_test
            y_test = self.y_test

        # calculate the error
        y_pred = self.ann.predict(X_test).flatten()
        y_pred = (y_pred>snr_threshold)
        y_test = (y_test>snr_threshold)

        # # Making the Confusion Matrix
        cm = confusion_matrix(y_test, y_pred)
        print(cm)

        accuracy = accuracy_score(y_test, y_pred)*100
        print(f"Accuracy: {accuracy:.3f}%")

        return cm, accuracy, y_pred, y_test

    def load_model_scaler_error(self, 
        ann_file_name='ann_model.h5',
        scaler_file_name='scaler.pkl',
        error_adjustment_file_name=False,
    ):
        # load the model
        self.ann = load_ann_h5(f'{self.directory}/{ann_file_name}')
        self.scaler = pickle.load(open(f'{self.directory}/{scaler_file_name}', 'rb'))
        if error_adjustment_file_name:
            self.error_adjustment = get_param_from_json(f'{self.directory}/{error_adjustment_file_name}')

            return self.ann, self.scaler, self.error_adjustment
        else:
            return self.ann, self.scaler

    def helper_error_adjustment(self, y_pred, y_test, snr_range=[4,10]):

        def linear_fit_fn(x, a, b):
            return a*x + b

        idx = (y_pred>snr_range[0]) & (y_pred<snr_range[1])
        idx &= (y_test != 0)
        popt, pcov = curve_fit(linear_fit_fn, y_pred[idx], y_pred[idx]-y_test[idx])

        adjustment_dict = {'slope': popt[0], 'intercept': popt[1]}

        return adjustment_dict

    def snr_error_adjustment(self, gw_param_dict=None, randomize=True, snr_threshold=8.0, snr_range=[4,10], error_adjustment_file_name='error_adjustment.json'):

        _, y_pred_, y_test_ = self.pdet_error(gw_param_dict=gw_param_dict, randomize=True, snr_threshold=snr_threshold)

        self.error_adjustment = self.helper_error_adjustment(y_pred_, y_test_, snr_range=snr_range)

        print(f"slope: {self.error_adjustment['slope']:.4f}, intercept: {self.error_adjustment['intercept']:.4f}")

        # save json file
        append_json(f'{self.directory}/{error_adjustment_file_name}', self.error_adjustment, replace=True)
        print(f"error adjustment saved at: {self.directory}/{error_adjustment_file_name}")

        return self.error_adjustment

    def predict_snr(self, gw_param_dict, error_adjustment=True):

        params = self.get_parameters(gw_param_dict)

        # input and output data
        X_test = self.get_input_data(params=params)
        X_test = self.scaler.transform(X_test)

        # calculate the error
        y_pred = self.ann.predict(X_test).flatten()

        # error adjustment
        if error_adjustment:
            adjustment_dict = self.error_adjustment
            a = adjustment_dict['slope']
            b = adjustment_dict['intercept']
            y_pred = y_pred-(a*y_pred + b)

        return y_pred

    def predict_pdet(self, gw_param_dict, snr_threshold=8.0, error_adjustment=True):

        y_pred = self.predict_snr(gw_param_dict, error_adjustment=error_adjustment)
        y_pred = (y_pred>snr_threshold)

        return y_pred